#include <iomanip>
#include "DataTypes.h"
#include "MeshLoader.h"


bool Node::operator==(const Node& right_node) const {
  return
    (id == right_node.id &&
      coord == right_node.coord &&
      top == right_node.top);
};

std::ostream& operator<<(std::ostream& out_stream, const Node& node_) {
  out_stream <<
    std::setw(4) << node_.id << "  " <<
    std::right << std::setw(11) <<
    node_.top << std::endl;
  for (const auto& elem : node_.coord) {
    out_stream << elem << " ";
  }

  return out_stream;
}

std::ostream& operator<<(std::ostream& out_stream, const FiniteElement& element_) {
  out_stream << std::setw(4) << element_.element_id << " |"
    << std::setw(11) << element_.material_id << " |";
  for (const auto& nodes_ID_it : element_.nodes_id)
    out_stream << std::right << std::setw(4) << nodes_ID_it;
  out_stream << std::endl;
  return out_stream;
}

std::ostream& operator<<(std::ostream& out_stream, const BoundaryFiniteElement& surface_) {
  out_stream << std::setw(9) << surface_.element_id << " |"
    << std::setw(9) << surface_.border_id << " |";
  for (const auto& nodes_ID_it : surface_.nodes_id)
    out_stream << std::right << std::setw(4) << nodes_ID_it;

  out_stream << std::endl;
  return out_stream;
}

std::ostream& operator<<(std::ostream& out_stream, const std::vector<Node>& nodes_) {
  out_stream << nodes_.size() << (nodes_.size() > 1 ? " Nodes:\n" : " Node:\n") << " ID"
    << std::setw(10) << "X" << std::setw(11) << "Y" << std::setw(11) << "Z" << std::setw(14) << "top" << std::endl;

  for (const auto& it : nodes_)
    out_stream << it;

  return out_stream;
}


Node::Node(int id_, std::array<double, 3> coord_, bool vertex_) 
  : id(id_)
  , coord(coord_)
  , top(vertex_) {}

bool Node::operator<(const Node& right_node) const {
  return id < right_node.id;
}

std::ostream& operator<<(std::ostream& out_stream, const std::vector<FiniteElement>& elements_) {
  int elements_size = elements_.size();
  out_stream << elements_size << (elements_size > 1 ? " FiniteElement:\n" : " FiniteElement:\n")
    << " ID" << std::setw(15) << "Material_ID" << std::setw(15) << "Nodes ID\n";

  for (const auto& it : elements_)
    out_stream << it;

  return out_stream;
}

std::ostream& operator<<(std::ostream& out_stream, const std::vector<BoundaryFiniteElement>& surfaces_) {
  int surfaces_size = surfaces_.size();
  out_stream << surfaces_size << (surfaces_size > 1 ? " BoundaryFiniteElement:\n" : " BoundaryFiniteElement:\n")
    << "Boundary_ID" << std::setw(11) << "Border_ID" << std::setw(11) << "Nodes ID\n";

  for (const auto& it : surfaces_)
    out_stream << it;

  return out_stream;
}

bool Edge::operator==(const Edge& right_edge) const {
  if (right_edge.first_node_id == first_node_id && right_edge.last_node_id == last_node_id)
    return true;
  else if (right_edge.first_node_id == last_node_id && right_edge.last_node_id == first_node_id)
    return true;
  else return false;
}

Edge::Edge(int id1, int id2, int id3) :
  first_node_id(id1), last_node_id(id2), center_id(id3) {
}

void Edge::UpdateCenter(int id_) {
  center_id = id_;
}

bool Edge::operator<(const Edge& right_edge) const {
  return center_id < right_edge.center_id;
}

std::size_t EdgeHash::operator() (const Edge& _p) const {
  std::size_t seed = 0;
  if (_p.first_node_id > _p.last_node_id) {
    seed ^= std::hash<int>() (_p.first_node_id) + 0x9e3779b9 + (seed << 6) + (seed >> 2);
    seed ^= std::hash<int>() (_p.last_node_id) + 0x9e3779b9 + (seed << 6) + (seed >> 2);
  } else {
    seed ^= std::hash<int>() (_p.last_node_id) + 0x9e3779b9 + (seed << 6) + (seed >> 2);
    seed ^= std::hash<int>() (_p.first_node_id) + 0x9e3779b9 + (seed << 6) + (seed >> 2);
  }

  return seed;
}
